# $Header: /usr/custom/cvs/public/EasyPublisher/lib/python/Products/EasyLanguageService/Attic/EasyLanguageService.py,v 1.1.2.3 2002/08/13 08:07:07 johanc Exp $
#
# (c) 2002, Torped Strategi och Kommunikation AB.
# See License.txt for License information.
#
__doc__='''EasyLanguageService product.'''
__version__='$Revision: 1.1.2.3 $'[11:-2]

import Globals
from types import ListType, TupleType, StringType
from Globals import DTMLFile
try: from Products.EasySecurity.EasyCorePermissions import *
except: EasyLanguageServiceManage = 'Easy LanguageService Manage'
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager

from AccessControl import Permissions
# General Zope permissions
View = Permissions.view
AccessContentsInformation = Permissions.access_contents_information

try:
    from Products.EasyNavigation.EasyNavigation import NavigationBrain
    import Acquisition
    import AccessControl.Role
    from ComputedAttribute import ComputedAttribute
    
    
    class LocalNavigationBrain(Acquisition.Implicit, NavigationBrain):  

         __allow_access_to_unprotected_subobjects__=1

         def __init__(self, dict={}):
             for key,item in dict.items():
                 setattr(self,key,item)

         def _physicalPath(self):
             af_url=self.absolute_folder_url
             if af_url[-1]=='/': af_url=af_url[:-1]
             if af_url[0]!='/': af_url='/'+af_url
             return tuple(af_url.split('/'))
         physicalPath = ComputedAttribute(_physicalPath, 1)

         def _virtualPath(self):
             root_len=len(self.getVirtualRootPath)
             path=self._physicalPath()
             if len(path)>=root_len:
                 path=('',) + path[root_len:]
             return path
         virtualPath = ComputedAttribute(_virtualPath, 1)

         def _physicalURL(self):
             return self.absolute_folder_url
         physicalURL = ComputedAttribute(_physicalURL, 1)

         def _virtualURL(self):
             vpath=self._virtualPath()
             return '/'.join(vpath) + '/'
         virtualURL = ComputedAttribute(_virtualURL, 1)

         def _virtualParentPath(self):
             return self._virtualPath()[:-1]
         virtualParentPath = ComputedAttribute(_virtualParentPath, 1)

         def _physicalParentPath(self):
             return self._physicalPath()[:-1]
         physicalParentPath = ComputedAttribute(_physicalParentPath, 1)

         def has_key(self, key):
            return hasattr(self, key)
                    
                    
except:
    NavigationBrain=None
    
    
EASYLANGUAGESERVICE='EasyLanguageService'

from rfc3066_language_codes import *

def isSequense(seq):
    return type(seq) in (ListType, TupleType)

def uniqueValues(seq):
    d={}
    map(d.setdefault, seq)
    return d.keys()


class EasyLanguageService(PropertyManager, SimpleItem):
    """
    The EasyLanguageService is use to keep information about
    language codes (RFS3066) and methods for choosing default 
    language (from browser default) an make translations to 
    human readable forms.
    
    Only ISO 639_1: 2-letter codes and combination of ISO 639: 2-letter codes
    and ISO 3166-1 country codes are currently supported. 
    This is what is supported by Internet Explorer and possibly other browsers.
    
    EasyLanguageService also helps a web site register supported
    language versions and the languages base url.
    
    """
    
    security = ClassSecurityInfo()
    
    meta_type = EASYLANGUAGESERVICE

    language_folders = {}
    language_mappings = {}

    preferred_languages='preferred_languages'

    _properties = PropertyManager._properties + \
            (   
                {'id':'preferred_languages', 'type': 'string', 'mode': 'w'},
            )

    manage_options_local =(
        {'label':'LanguageFolders', 'action':'manage_main'},
        {'label':'LanguageMappings', 'action':'manage_mappings'},
    )

    manage_options_property_manager = PropertyManager.manage_options
    manage_options_simple_item = SimpleItem.manage_options

    manage_options = manage_options_local + manage_options_property_manager + manage_options_simple_item

    def __init__(self, id, title=EASYLANGUAGESERVICE):
        self.id=id
        self.title=title

    #Query Interface
    
    security.declarePublic('getPreferredLanguagesName')
    def getPreferredLanguagesName(self):
        """Returns the name of the preferred lanugage request attribute."""
        return self.preferred_languages
    
    security.declarePublic('getPreferredLanguages')
    def getPreferredLanguages(self, object=None, use_substitute=1):
        """Returns the language code that would be most like
        preferred by the user.
        """
        session_lang=self.getPreferredLanguageInSession([])
        if session_lang!=[]:
            session_lang=[session_lang]
        obj_pl = getattr(object,self.preferred_languages,[])
        if type(obj_pl)==StringType: obj_pl=[obj_pl]
        REQUEST = getattr(self,'REQUEST', {})
        req_pl = REQUEST.get(self.preferred_languages, [])
        if type(req_pl)==StringType: req_pl=[req_pl]
        brw_pl_str = REQUEST.get('HTTP_ACCEPT_LANGUAGE', '')
        brw_pl =  brw_pl_str.split(',')
        brw_pl = map(lambda x: x.split(';')[0], brw_pl)
        if use_substitute:
            return self.getSubstitute(req_pl + session_lang + obj_pl + brw_pl)
        return req_pl + session_lang + obj_pl + brw_pl
   
    security.declarePublic('getSubstitute')
    def getSubstitute(self, languages=[]):
        values=[]
        for lang in languages:
            values.append(self.language_mappings.get(lang, lang))
        return values

    security.declarePublic('filterAllowed')
    def filterAllowed(self, languages=[], default=['en']):
        """Filter languages that are defined with home pages (or none)"""
        filtered=[]
        append=filtered.append
        allowed = self.language_folders.keys()
        for lang in languages:
            if lang in allowed: append(lang)
        if not filtered: return default
        return filtered

    security.declarePublic('getPreferredLanguages')
    def getAllowedPreferredLanguages(self, object=None, use_substitute=1, default=['en']):
        """Return allowed preferred languages."""
        pl=self.getPreferredLanguages(object, use_substitute)
        return self.filterAllowed(pl, default)

    security.declarePublic('getPreferredLanguageCode')
    def getPreferredLanguageCode(self, object=None, use_substitute=1, default='en'):
        """Return allowed preferred language."""
        lang=self.getAllowedPreferredLanguages(object=object,use_substitute=use_substitute,default=default)
        if len(lang)>0: return lang[0]
        return default

    security.declarePublic('getPreferredLanguage')
    def getPreferredLanguage(self, object=None, use_substitute=1, default='English'):
        """Return allowed preferred language as plain text (English)."""
        language=self.getLanguageByRFC3066(self.getPreferredLanguageCode(object=object,use_substitute=use_substitute), None)
        if language is not None: 
            return language
        return default
        
    security.declarePrivate('getFolder')
    def getFolder(self, lang, default=None):
        af_url = self.language_folders.get(lang,default)['absolute_folder_url']
        if af_url:
            try:
                return self.restrictedTraverse(af_url)
            except: pass
        return default

    security.declarePublic('getPreferredFolder')
    def getPreferredFolder(self, object=None, use_substitute=1, default=None):
        """Return the preferred languages folder object."""
        allowed = self.getAllowedPreferredLanguages(object, use_substitute)
        folder=None
        for lang in allowed:
            folder=self.getFolder(lang, None)
            if folder is not None: break;
        if folder is None: return default
        return folder

    security.declarePublic('getPreferredFolderAbsoluteURL')
    def getPreferredFolderAbsoluteURL(self, object=None, use_substitute=1, default=None):
        """Return the absolute URL to preferred languages folder."""
        allowed = self.getAllowedPreferredLanguages(object, use_substitute)
        url=None
        for lang in allowed:
            url = self.language_folders.get(lang,{}).get('absolute_folder_url',None)
            if url is not None: break;
        if url is None: return default
        if url[-1]!='/': url=url+'/'
        return url

    security.declarePublic('getPreferredFolderURL')
    def getPreferredFolderURL(self, object=None, use_substitute=1, default=None):
        """Return the virtual URL to preferred languages folder."""
        url = self.getPreferredFolderAbsoluteURL(object, use_substitute, default)
        if url is not None: 
            url = self.relativeUrl(url)        
            return url
        return default


    # Action Interface

    security.declarePublic('redirect')
    def redirect(self, object=None, use_substitute=1, default=None, lock=1):
        """Redirect to the Language Folder, and ignore exceptions that may be raise after this."""
        location=self.getPreferredFolderURL(object=object, use_substitute=use_substitute, default=default)
        if location:
            self.REQUEST.RESPONSE.redirect(location, lock=lock)


    # Language Query Interface
    
    security.declarePublic('getAllLanguage')
    def getAllLanguage(self, filter=None):
        """
        Returns all languages as (RFC3066, PlainText) except for the once in filter.
        
        filter is a sequence of RFC3066 code.
        """
        return getAllLanguage(filter)


    # Session Interface    
    security.declarePublic('setPreferredLanguageInSession')
    def setPreferredLanguageInSession(self, language=None, REQUEST=None):
        """Sets a cookie with the preferred language."""
        if language and validRFC3066(language):
            sdm = getattr(self, 'session_data_manager', None)
            if sdm:
                sd=sdm.getSessionData(create=1)
                sd.set(self.preferred_languages, language)
        if REQUEST:
            return "<pre>Language set to: '%s' in session variable '%s'.</pre>" % (language,self.preferred_languages)
            
    security.declarePublic('removePreferredLanguageInSession')
    def removePreferredLanguageInSession(self, REQUEST=None):
        """Sets a cookie with the preferred language."""
        sdm = getattr(self, 'session_data_manager', None)
        if sdm:
            sd=sdm.getSessionData(create=1)
            sd.delete(self.preferred_languages)
        if REQUEST:
            return "<pre>Language deleted in session variable '%s'.</pre>" % (language,self.preferred_languages)


    security.declarePublic('getPreferredLanguageInSession')
    def getPreferredLanguageInSession(self, default='en', REQUEST=None):
        """Gets the cookie with the preferred language or default."""
        sdm = getattr(self, 'session_data_manager', None)
        if sdm:
            sd=sdm.getSessionData(create=1)
            language = sd.get(self.preferred_languages, default)
        else:
            language = default
        if REQUEST:
             return "<pre>Language in session variable '%s' is set to: '%s' (default: '%s').</pre>" % (self.preferred_languages, language, default)
        return language
        
        
    # RFC/ISO Query Interface

    security.declarePublic('getLanguageByRFC3066')
    def getLanguageByRFC3066(self, code, default='English'):
        """Translates a language code to a human readable form in english.
        
        If a sequence of language codes is passed, a sequence of the corresponding
        languages is returns.
        """
        return getLanguageByRFC3066(code, default)

    security.declarePublic('getLanguageByISO639_1')
    def getLanguageByISO639_1(self, code, default='English'):
        """Translates a language code to a human readable form in english.
        
        If a sequence of language codes is passed, a sequence of the corresponding
        languages is returns.
        """
        return getLanguageByISO639_1(code, default)

    security.declarePublic('getCountryByISO3166_1')
    def getCountryByISO3166_1(self, code, default='United States'):
        """Translates a language code to a human readable form in english.
        
        If a sequence of language codes is passed, a sequence of the corresponding
        languages is returns.
        """
        return getCountryByISO3166_1(code, default)

    security.declarePublic('getRFC3066ByLanguage')
    def getRFC3066ByLanguage(self, language, default='en'):
        """Translates a language in a human readable form to the code.
        """
        return getRFC3066ByLanguage(language, default)

    security.declarePublic('getISO639_1ByLanguage')
    def getISO639_1ByLanguage(self, language, default='en'):
        """Translates a language in a human readable form to the code."""
        return getISO639_1ByLanguage(language, default)
        
    security.declarePublic('getISO3166_1ByCountry')
    def getISO3166_1ByCountry(self, country, default='en'):
        """Translates a language in a human readable form to the code."""
        return getISO3166_1ByCountry(country, default)

    security.declarePublic('getISO639_1ByRFC3066')
    def getISO639_1ByRFC3066(self, code):
        """
        Returns the root language class of the language code.
        
        If a sequence of language codes is passed, a tuple of the corresponding
        languages is returns.        
        """
        return getISO639_1ByRFC3066(code) 

    security.declarePublic('validISO639_1')
    def validISO639_1(self, code):
        """
        Returnes true if this is a valid language code.
        """
        return validISO639_1(code)

    security.declarePublic('validRFC3066')
    def validRFC3066(self, code):
        """
        Returnes true if this is a valid language code.
        """
        return validRFC3066(code)

    security.declarePublic('validISO3166_1')
    def validISO3166_1(self, code):
        """
        Returnes true if this is a valid language code.
        """
        return validISO3166_1(code)

    _manage_main = DTMLFile('dtml/manage_main', globals())
    security.declareProtected(EasyLanguageServiceManage, 'manage_main')
    def manage_main(self
                    , selected_language=[]
                    , add_language=None
                    , add_folder_url=None
                    , action='View'
                    , REQUEST=None):
        """Manage Language Folders"""
        management_view='LanguageFolders'
        manage_tabs_message=""
        deleted_languages=[]
        if action=='Delete' and selected_language:
            language_folders=self.language_folders
            for key in selected_language:
                del language_folders[key]
            self.language_folders=language_folders
            manage_tabs_message="Language(s) %s [%s] was deleted." % (getLanguageByRFC3066(selected_language), selected_language)
            deleted_languages=selected_language[0:1]
        elif action=='Add Language' and add_language is not None:
            changed_or_added='added'
            if add_folder_url!='':
                #ADD or CHANGE languages that do defined a language folder.
                root=REQUEST.get('VirtualRootPhysicalPath', ())
                language_folders=self.language_folders
                if language_folders.has_key(add_language): changed_or_added='changed'
                if add_folder_url[0]!='/': add_folder_url='/'+add_folder_url
                add_folder_url = '/'.join(root) + add_folder_url            
                physical = self.getPhysicalPath()
                parent_url = '/'.join(physical[:-1])
                if len(add_folder_url)>len(parent_url) and parent_url==add_folder_url[:len(parent_url)]: 
                    try:
                        self.restrictedTraverse(add_folder_url)
                        d = {
                                'absolute_folder_url':add_folder_url,
                                'rfc3066':add_language,
                                'language':getLanguageByRFC3066(add_language)
                            }
                        language_folders[add_language] = d
                        manage_tabs_message="Language %s was %s with folder %s." % (getLanguageByRFC3066(add_language), changed_or_added, self.relativeUrl(add_folder_url))
                        self.language_folders=language_folders
                    except:
                        manage_tabs_message="Failed! Invalid url %s, language %s was not %s." % (self.relativeUrl(add_folder_url),getLanguageByRFC3066(add_language),changed_or_added)
                else:
                    manage_tabs_message="Failed! You can't select a folder that isn't under the parent of this service."
            elif add_folder_url=='':
                #ADD or CHANGE languages that don't defined a language folder.
                changed_or_added='added'
                language_folders=self.language_folders                    
                if language_folders.has_key(add_language): changed_or_added='changed'
                d = {
                        'absolute_folder_url':add_folder_url,
                        'rfc3066':add_language,
                        'language':getLanguageByRFC3066(add_language)
                    }  
                language_folders[add_language] = d
                manage_tabs_message="Language %s was %s." % (getLanguageByRFC3066(add_language), changed_or_added)
                self.language_folders=language_folders        
        return self._manage_main(self, REQUEST=REQUEST
                                , manage_tabs_message=manage_tabs_message
                                , management_view=management_view
                                , deleted_languages=deleted_languages
                                )

    
    _manage_mappings = DTMLFile('dtml/manage_mappings', globals())
    security.declareProtected(EasyLanguageServiceManage, 'manage_mappings')
    def manage_mappings(self
                    , selected_mappings=[]
                    , add_mapping=None
                    , add_languages=[]
                    , action='View'
                    , REQUEST=None):
        """Manage Language Folders"""
        management_view='LanguageFolders'
        manage_tabs_message=""
        deleted_mapping=None
        deleted_languages=None
        keys_in_use = self.getLanguageMappingUsedKeys()
        
        if action=='Delete' and selected_mappings:
            language_mappings=self.language_mappings
            if len(selected_mappings)==1:
                deleted_mapping=selected_mappings[0]
                deleted_languages=[]
            for item, key in language_mappings.items():
                if key in selected_mappings:
                    del language_mappings[item]
                    if len(selected_mappings)==1:
                        deleted_languages.append(item)
            self.language_mappings=language_mappings
            manage_tabs_message="Mapping to language(s) %s [%s] was deleted." % (', '.join(getLanguageByRFC3066(selected_mappings)), ', '.join(selected_mappings))

        elif action=='Add Mapping' and add_mapping and add_mapping not in keys_in_use:
            language_mappings=self.language_mappings
            keys_in_use.append(add_mapping)
            add_languages = filter(lambda x,keys_in_use=keys_in_use: x not in keys_in_use, add_languages)
            for lang in add_languages:
                language_mappings[lang] = add_mapping
            self.language_mappings=language_mappings

        return self._manage_mappings(self,REQUEST=REQUEST
                                    , manage_tabs_message=manage_tabs_message
                                    , management_view=management_view
                                    , deleted_mapping=deleted_mapping
                                    , deleted_languages=deleted_languages
                                    )


    security.declarePublic('relativeUrl')
    def relativeUrl(self, absolute_url, REQUEST=None):  
        if REQUEST is None: REQUEST=self.REQUEST
        root_path=REQUEST.get('VirtualRootPhysicalPath', ())
        root_url='/'.join(root_path)
        return absolute_url[len(root_url):]

    security.declarePublic('getLanguageFolderItems')
    def getLanguageFolderItems(self,spec=None):
        return self.language_folders.items()

    security.declarePublic('getLanguageFolderKeys')
    def getLanguageFolderKeys(self, spec=[]):
        return self.language_folders.keys() + spec
        
    security.declarePublic('getLanguageMappingItems')
    def getLanguageMappingItems(self,spec=None):
        values = {}
        for key,item in self.language_mappings.items():
            new_item=values.get(item, [])
            new_item.append(key)
            values[item]=new_item
        return values.items()

    security.declarePublic('getLanguageMappingKeys')
    def getLanguageMappingKeys(self,spec=None):
        return self.language_mappings.values()

    security.declarePublic('getLanguageMappingUsedKeys')
    def getLanguageMappingUsedKeys(self,spec=[]):
        """Returns all unique ids that has a mapping or is mapped from."""
        result={}
        set=result.setdefault
        if spec:
            map(lambda x, set=set: set(x), spec)
        keys=self.language_mappings.keys()
        values=self.language_mappings.values()
        if keys:
            map(lambda x, set=set: set(x), keys)
        if values:
            map(lambda x, set=set: set(x), values)           
        return result.keys()
        
    security.declarePublic('getLanguageMappedKeys')
    def getLanguageMappedKeys(self):
        """Returns all unique ids that is mapping to another language."""
        return self.language_mappings.keys()       
        
    security.declarePublic('getLanguageItems')
    def getLanguageItems(self, spec=[]):
        """Returns all languages as (rfc3066Code, PlainText)"""
        return getLanguageItems(spec=spec)

    
    #Navigation API

    if NavigationBrain:
        security.declarePublic(View, 'getMenu')
        def getMenu(self, object_path, sort_on='order', sort_order=None, skip_root=None):
            brains = []
            append=brains.append
            vrpp=self.REQUEST.get('VirtualRootPhysicalPath', ())
            navigation_menu = self.getPhysicalPath()
            if hasattr(object_path,'physicalPath'): object_path=object_path.physicalPath
                    
            LocalNavigationBrain.query=None
            LocalNavigationBrain.object_path=object_path
            LocalNavigationBrain.skip_root=skip_root
            LocalNavigationBrain.getMenu=None
            LocalNavigationBrain.virtual_root_path_of_request=vrpp
            LocalNavigationBrain.NavigationService=self        
            LocalNavigationBrain.isAnObjectManager=1  
            LocalNavigationBrain.getVirtualRootPath=vrpp
            LocalNavigationBrain.navigation_menus=self.title or self.id
            
            for key, item in self.getLanguageFolderItems():
                brains.append(LocalNavigationBrain(dict={
                              'absolute_folder_url':    item['absolute_folder_url']
                            , 'rfc3066':                item['rfc3066']
                            , 'language':               item['language']
                            , 'navigation_title':       item['language']
                            , 'title':                  item['language']
                            , 'description':            item['language']
                        }).__of__(self))

            #Post sort
            #def sortmethod( x,y, nav_menu):
            #    return cmp(x.navigation_menus_data[nav_menu][0],
            #               y.navigation_menus_data[nav_menu][0])
            #
            #if sort_on=='order':
            #    menuobjectlist = list(menuobjectlist)
            #    menuobjectlist.sort( lambda x,y, nav_menu=object_path,
            #                     sortmethod=sortmethod: sortmethod(x,y,nav_menu))
            #    if sort_order=='reverse':
            #        menuobjectlist.reverse()
        
            return brains
    
Globals.InitializeClass(EasyLanguageService)



def getLanguageItems(spec=[]):
    """
    Returns all languages as (RFC3066, PlainText) except for the once in filter.
        filter is a sequence of RFC3066 code.
    """

    keys=rfc_3066.keys()
    keys=filter(lambda x, spec=spec: x not in spec, keys)
    keys.sort()
    return map(lambda key, rfc_3066=rfc_3066: (key, rfc_3066[key]) , keys)


def getLanguageByRFC3066(code, default='English'):
    """Translates a language code to a human readable form in english.

    If a sequence of language codes is passed, a sequence of the corresponding
    languages is returns.
    """
    get=rfc_3066.get
    if isSequense(code):
        return map(lambda x, d=default, get=get: get(x,d), code)
    else:
        return get(code, default)

def getLanguageByISO639_1(code, default='English'):
    """Translates a language code to a human readable form in english.

    If a sequence of language codes is passed, a sequence of the corresponding
    languages is returns.
    """
    get=iso_639_1.get
    if isSequense(code):
        return map(lambda x, d=default, get=get: get(x[:2],d), code)
    else:
        return get(code[:2], default)


def getCountryByISO3166_1(code, default='United States'):
    """Translates a language code to a human readable form in english.

    If a sequence of language codes is passed, a sequence of the corresponding
    languages is returns.
    """
    get=iso_3166_1.get
    if isSequense(code):
        return map(lambda x, d=default, get=get: get(x,d), code)
    else:
        return get(code, default)

def getRFC3066ByLanguage(language, default='en'):
    """Translates a language in a human readable form to the code.
    """
    # 1. Should be a re.search or something?
    # 2. default not in use!!
    return filter(lambda k,v,language=language: v==language, rfc_3066.items())

def getISO639_1ByLanguage(language, default='en'):
    """Translates a language in a human readable form to the code."""
    # 1. Should be a re.search or something?
    # 2. default not in use!!
    return filter(lambda k,v,language=language: v==language, iso_639_1.items())

def getISO3166_1ByCountry(country, default='en'):
    """Translates a language in a human readable form to the code."""
    # 1. Should be a re.search or something?
    # 2. default not in use!!
    return filter(lambda k,v,country=country: v==country, iso_3166_1.items())

def getISO639_1ByRFC3066(code):
    """
    Returns the root language class of the language code.

    If a sequence of language codes is passed, a tuple of the corresponding
    languages is returns.        
    """
    if isSequense(code):
        d={}
        map(d.setdefault, map(lambda x: x[:2], code))
        return d.keys()
    else:
        return code[:2]   

def validISO639_1(code):
    """
    Returnes true if this is a valid language code.
    """
    haskey=iso_639_1.has_key
    if isSequense(code):
        return map(lambda x, haskey=haskey: haskey(x), code)
    else:        
        return haskey(code)

def validRFC3066(code):
    """
    Returnes true if this is a valid language code.
    """
    haskey=rfc_3066.has_key
    if isSequense(code):
        return map(lambda x, haskey=haskey: haskey(x), code)
    else:
        return haskey(code)

def validISO3166_1(code):
    """
    Returnes true if this is a valid language code.
    """
    haskey=iso_3166_1.has_key
    if isSequense(code):
        return map(lambda x, haskey=has_key: haskey(x), code)
    else:        
        return haskey(code)

def manage_addEasyLanguageService(self, title='', REQUEST=None):
    """Add a EasyLanguageService into a web site."""

    id=EASYLANGUAGESERVICE

    ob=EasyLanguageService(id)
    self._setObject(id,ob)  
    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_main')



def initialize(context):
    """Initialize EasyMailService Product"""


    context.registerClass(
        EasyLanguageService,
        permission=EasyLanguageServiceManage,
        constructors=(manage_addEasyLanguageService,
                      manage_addEasyLanguageService),
        icon='icons/EasyLanguageService.gif',
    )


